searchData={"content_type":"text/markdown","items":[{"doc":"use CharonOauth2, @charon_config","ref":"CharonOauth2.html","title":"CharonOauth2","type":"module"},{"doc":"Default token verification call used in refresh_token grant.\nCan be overridden using config opt `:verify_refresh_token`,\nin which case there is no need to call `Charon.TokenPlugs.verify_no_auth_error/2`.\n\nChecks:\n - signature valid\n - already valid (\"nbf\" claim)\n - not expired (\"exp\" claim)\n - token type is \"refresh\" (\"type\" claim)\n - session type is \"oauth2\" (\"styp\" claim)\n - session exists\n - token is fresh (grace period 10 seconds)","ref":"CharonOauth2.html#verify_refresh_token/2","title":"CharonOauth2.verify_refresh_token/2","type":"function"},{"doc":"Config module for `CharonOauth2`.\n\nUnlike `Charon` itself, not all config is runtime config.\nThat means overriding some configuration options at runtime may not result in the expected behaviour.\nThe reason for this is that several configuration values are read at compile time in order to generate code using macros.\nThese config options are also read by the migration helper `CharonOauth2.Migration`,\nand you should generally not change them after initializing CharonOauth2.\nThe glossary below specifies which config options are affected.\n\n    Charon.Config.from_enum(\n      ...,\n      optional_modules: %{\n        CharonOauth2 => %{\n          repo: MyApp.Repo,\n          resource_owner_schema: MyApp.User,\n          scopes: ~w(profile:read door:open),\n          # following are defaults\n          authorizations_table: \"charon_oauth2_authorizations\",\n          clients_table: \"charon_oauth2_clients\",\n          customize_session_upsert_args: &Function.identity/1,\n          enforce_pkce: :all,\n          grants_table: \"charon_oauth2_grants\",\n          grant_ttl: 10 * 60,\n          resource_owner_id_column: :id,\n          resource_owner_id_type: :bigserial,\n          test_seed_defaults: %{}\n          verify_refresh_token: &CharonOauth2.verify_refresh_token/2,\n        }\n      }\n    )","ref":"CharonOauth2.Config.html","title":"CharonOauth2.Config","type":"module"},{"doc":"- `:authorizations_table` (compile-time) the name of the table in which to store authorizations.\n - `:clients_table` (compile-time) the name of the table in which to store clients.\n - `:customize_session_upsert_args` a function that you can use to customize the arguments that are passed by your `MyApp.TokenEndpoint` to `Charon.SessionPlugs.upsert_session/3`. Be careful, usually you might want to add to these arguments, but not override them.\n - `:enforce_pkce` for `:public`, `:all` or `:no` clients\n - `:grant_ttl` time in seconds that a grant (mostly authorization code) takes to expire\n - `:grants_table` (compile-time) the name of the table in which to store grants\n - `:repo` (required, compile-time) the Ecto repo module of your application.\n - `:resource_owner_id_column` (compile-time) the column name, as an atom, of the resource owner's schema's primary key\n - `:resource_owner_id_type` (compile-time) the type, as an atom, of the resource owner's schema's primary key\n - `:resource_owner_schema` (required, compile-time) the user schema module of your application.\n - `:resource_owner_table` (compile-time) the name of the table in resource owners are stored. Taken from `:resource_owner_schema` unless set.\n - `:scopes` (required, compile time) the scopes that are available to Oauth2 apps, application-wide.\n - `:test_seed_defaults` override default values for test models used in `CharonOauth2.TestSeeds`.\n - `:verify_refresh_token` a function that you can use to verify an Oauth2 refresh token for the refresh token grant.","ref":"CharonOauth2.Config.html#module-glossary","title":"Glossary - CharonOauth2.Config","type":"module"},{"doc":"Build config struct from enumerable (useful for passing in application environment).\nRaises for missing mandatory keys and sets defaults for optional keys.","ref":"CharonOauth2.Config.html#from_enum/1","title":"CharonOauth2.Config.from_enum/1","type":"function"},{"doc":"","ref":"CharonOauth2.Config.html#t:t/0","title":"CharonOauth2.Config.t/0","type":"type"},{"doc":"Helper module for client app migrations.","ref":"CharonOauth2.Migration.html","title":"CharonOauth2.Migration","type":"module"},{"doc":"defmodule MyApp.Repo.Migrations.Oauth2Models do\n      use Ecto.Migration\n\n      def change, do: CharonOauth2.Migration.change(charon_config)\n    end","ref":"CharonOauth2.Migration.html#module-usage","title":"Usage - CharonOauth2.Migration","type":"module"},{"doc":"Call from a migration to generate `CharonOauth2` models. Supports rollbacks as well.","ref":"CharonOauth2.Migration.html#change/1","title":"CharonOauth2.Migration.change/1","type":"function"},{"doc":"Ecto type for encrypted data. Takes any binary as input.\n\nRequires a column of type `:binary`.","ref":"CharonOauth2.Types.Encrypted.html","title":"CharonOauth2.Types.Encrypted","type":"module"},{"doc":"Ecto type to only store an HMAC of the input data.\nThis means the original data cannot be retrieved,\nbut a record can be found by an exact equality match.\n\nRequires a column of type `:binary`.","ref":"CharonOauth2.Types.Hmac.html","title":"CharonOauth2.Types.Hmac","type":"module"},{"doc":"Ecto type for casting a string with a pattern-separated list of values to an ordered-set string array.\nAll pattern types of `String.split/3` are supported.\nEmpty strings are trimmed from the result by default.","ref":"CharonOauth2.Types.SeparatedStringOrdset.html","title":"CharonOauth2.Types.SeparatedStringOrdset","type":"module"},{"doc":"schema \"separated_things\" do\n      field :comma_separated, SeparatedStringOrdset, pattern: \",\"\n      field :comma_or_semicolon_separated, SeparatedStringOrdset, pattern: ~w(, ;)\n      field :regex_separated, SeparatedStringOrdset, pattern: ~r/,/\n      field :comma_separated_with_empty_strings, SeparatedStringOrdset, pattern: \",\", split_opts: []\n    end","ref":"CharonOauth2.Types.SeparatedStringOrdset.html#module-example","title":"Example - CharonOauth2.Types.SeparatedStringOrdset","type":"module"},{"doc":"@opts SeparatedStringOrdset.init(pattern: \",\")\n\n    iex> {:ok, ~w(bar baz foo)} = SeparatedStringOrdset.cast(\"foo,bar,baz,bar\", @opts)\n    iex> :error = SeparatedStringOrdset.cast([1, 2, \"3\"], @opts)\n    iex> {:ok, ~w(1 2 true)} = SeparatedStringOrdset.cast(\"1,2,true\", @opts)\n    iex> {:ok, ~w(a b c)} = SeparatedStringOrdset.cast(~w(a b c), @opts)\n\n    iex> opts = SeparatedStringOrdset.init(pattern: \",\", split_opts: [])\n    iex> {:ok, [\"\", \"bar\", \"foo\"]} = SeparatedStringOrdset.cast(\"foo,bar,\", opts)","ref":"CharonOauth2.Types.SeparatedStringOrdset.html#module-doctests","title":"Doctests - CharonOauth2.Types.SeparatedStringOrdset","type":"module"},{"doc":"A dummy app used to test `CharonOauth2` and to present documentation for the generated modules.","ref":"MyApp.html","title":"MyApp","type":"module"},{"doc":"Entrypoint module for CharonOauth2.\n\nThe following submodules are generated:\n- `Elixir.MyApp.CharonOauth2.Authorization`\n- `Elixir.MyApp.CharonOauth2.Authorizations`\n - `Elixir.MyApp.CharonOauth2.Client`\n - `Elixir.MyApp.CharonOauth2.Clients`\n - `Elixir.MyApp.CharonOauth2.Grant`\n - `Elixir.MyApp.CharonOauth2.Grants`\n - `Elixir.MyApp.CharonOauth2.Plugs.AuthorizationEndpoint`\n - `Elixir.MyApp.CharonOauth2.Plugs.TokenEndpoint`\n - `Elixir.MyApp.CharonOauth2.TestSeeds`","ref":"MyApp.CharonOauth2.html","title":"MyApp.CharonOauth2","type":"module"},{"doc":"An authorization represents the permission granted by a resource owner (usually a user)\nto an application to act on their behalf within certain limits (determined by scopes).\n\nField `:scope` is guaranteed to be an ordset (`:ordsets`).","ref":"MyApp.CharonOauth2.Authorization.html","title":"MyApp.CharonOauth2.Authorization","type":"module"},{"doc":"Basic changeset.","ref":"MyApp.CharonOauth2.Authorization.html#changeset/2","title":"MyApp.CharonOauth2.Authorization.changeset/2","type":"function"},{"doc":"Insert-only changeset - some things (should) never change.","ref":"MyApp.CharonOauth2.Authorization.html#insert_only_changeset/2","title":"MyApp.CharonOauth2.Authorization.insert_only_changeset/2","type":"function"},{"doc":"Returns a new query with the current module as named binding `:charon_oauth2_authorization`.","ref":"MyApp.CharonOauth2.Authorization.html#named_binding/0","title":"MyApp.CharonOauth2.Authorization.named_binding/0","type":"function"},{"doc":"Preload named bindings. Automatically joins using `resolve_binding/2`.","ref":"MyApp.CharonOauth2.Authorization.html#preload/2","title":"MyApp.CharonOauth2.Authorization.preload/2","type":"function"},{"doc":"Resolve named bindings that are not present in the query by (left-)joining to the appropriate tables.","ref":"MyApp.CharonOauth2.Authorization.html#resolve_binding/2","title":"MyApp.CharonOauth2.Authorization.resolve_binding/2","type":"function"},{"doc":"Bindings / preloads that can be used with `resolve_binding/2` and `preload/2`","ref":"MyApp.CharonOauth2.Authorization.html#t:resolvable/0","title":"MyApp.CharonOauth2.Authorization.resolvable/0","type":"type"},{"doc":"","ref":"MyApp.CharonOauth2.Authorization.html#t:t/0","title":"MyApp.CharonOauth2.Authorization.t/0","type":"type"},{"doc":"Context to manage authorizations","ref":"MyApp.CharonOauth2.Authorizations.html","title":"MyApp.CharonOauth2.Authorizations","type":"module"},{"doc":"Get a list of all oauth2 authorizations.","ref":"MyApp.CharonOauth2.Authorizations.html#all/2","title":"MyApp.CharonOauth2.Authorizations.all/2","type":"function"},{"doc":"iex> insert_test_authorization!(resource_owner_id: insert_test_user().id)\n    iex> [%Authorization{}] = Authorizations.all()\n\n    # can be filtered\n    iex> authorization = insert_test_authorization!(resource_owner_id: insert_test_user().id)\n    iex> [%Authorization{}] = Authorizations.all(%{resource_owner_id: authorization.resource_owner_id})\n    iex> [%Authorization{}] = Authorizations.all(%{scope: authorization.scope |> List.first()})\n    iex> [] = Authorizations.all(%{resource_owner_id: authorization.resource_owner_id + 1})","ref":"MyApp.CharonOauth2.Authorizations.html#all/2-doctests","title":"Doctests - MyApp.CharonOauth2.Authorizations.all/2","type":"function"},{"doc":"Delete an authorization.","ref":"MyApp.CharonOauth2.Authorizations.html#delete/1","title":"MyApp.CharonOauth2.Authorizations.delete/1","type":"function"},{"doc":"# authorization must exist\n    iex> {:error, :not_found} = Authorizations.delete(id: -1)\n\n    # succesfully deletes an authorization\n    iex> authorization = insert_test_authorization!(resource_owner_id: insert_test_user().id)\n    iex> {:ok, _} = Authorizations.delete([id: authorization.id])\n    iex> {:error, :not_found} = Authorizations.delete([id: authorization.id])","ref":"MyApp.CharonOauth2.Authorizations.html#delete/1-examples-doctests","title":"Examples / doctests - MyApp.CharonOauth2.Authorizations.delete/1","type":"function"},{"doc":"Get a single authorization by one or more clauses, optionally with preloads.\nReturns nil if Authorization cannot be found.","ref":"MyApp.CharonOauth2.Authorizations.html#get_by/2","title":"MyApp.CharonOauth2.Authorizations.get_by/2","type":"function"},{"doc":"iex> authorization = insert_test_authorization!(resource_owner_id: insert_test_user().id)\n    iex> %Authorization{} = Authorizations.get_by(id: authorization.id)\n    iex> nil = Authorizations.get_by(id: authorization.id + 1)\n\n    # preloads things\n    iex> authorization = insert_test_authorization!(resource_owner_id: insert_test_user().id)\n    iex> %{resource_owner: %{id: _}, client: %{id: _}} = Authorizations.get_by([id: authorization.id], Authorization.supported_preloads)","ref":"MyApp.CharonOauth2.Authorizations.html#get_by/2-doctests","title":"Doctests - MyApp.CharonOauth2.Authorizations.get_by/2","type":"function"},{"doc":"Insert a new authorization","ref":"MyApp.CharonOauth2.Authorizations.html#insert/1","title":"MyApp.CharonOauth2.Authorizations.insert/1","type":"function"},{"doc":"# succesfully creates an authorization\n    iex> {:ok, _} = insert_test_authorization(resource_owner_id: insert_test_user().id)\n\n    # a user can authorize a client only once\n    iex> {:ok, authorization} = insert_test_authorization(resource_owner_id: insert_test_user().id)\n    iex> insert_test_authorization(resource_owner_id: authorization.resource_owner_id, client_id: authorization.client_id) |> errors_on()\n    %{client_id: [\"user already authorized this client\"]}\n\n    # owner and client must exist\n    iex> insert_test_authorization(resource_owner_id: -1, client_id: insert_test_client!(owner_id: insert_test_user().id).id) |> errors_on()\n    %{resource_owner: [\"does not exist\"]}\n    iex> insert_test_authorization(resource_owner_id: insert_test_user().id, client_id: Ecto.UUID.generate()) |> errors_on()\n    %{client: [\"does not exist\"]}\n\n    iex> Authorizations.insert(%{}) |> errors_on()\n    %{scope: [\"can't be blank\"], client_id: [\"can't be blank\"], resource_owner_id: [\"can't be blank\"]}","ref":"MyApp.CharonOauth2.Authorizations.html#insert/1-examples-doctests","title":"Examples / doctests - MyApp.CharonOauth2.Authorizations.insert/1","type":"function"},{"doc":"Update an authorization.\n\n# TODO: fix ~(write write) test by implementing overrides","ref":"MyApp.CharonOauth2.Authorizations.html#update/2","title":"MyApp.CharonOauth2.Authorizations.update/2","type":"function"},{"doc":"# scopes must be subset of configured scopes and of client scopes\n    iex> insert_test_authorization!(resource_owner_id: insert_test_user().id) |> Authorizations.update(%{scope: ~w(cry)}) |> errors_on()\n    %{scope: [\"must be subset of party, read, write\"]}\n    iex> insert_test_authorization!(resource_owner_id: insert_test_user().id) |> Authorizations.update(%{scope: ~w(write write)}) |> errors_on()\n    %{scope: [\"client not allowed to access scope(s): write\"]}\n\n    # client and resource owner can't be updated\n    iex> %{id: id, client_id: client_id, resource_owner_id: owner_id} = insert_test_authorization!(resource_owner_id: insert_test_user().id)\n    iex> {:ok, %{client_id: ^client_id, resource_owner_id: ^owner_id}} = Authorizations.update([id: id], %{client_id: -1, owner_id: -1})","ref":"MyApp.CharonOauth2.Authorizations.html#update/2-examples-doctests","title":"Examples / doctests - MyApp.CharonOauth2.Authorizations.update/2","type":"function"},{"doc":"An Oauth2 (third-party) client application.\n\nFields `:scope`, `:grant_types` and `:redirect_uris` are guaranteed to be ordsets (`:ordsets`).\n\nThe client secret is stored encrypted, not hashed, which is unusual for passwords-like secrets.\nIn line with, for example, GCP, we choose to be able to show the client secret again.\nAlso, the client secret is not a normal password but a 384 bits random string,\nwhich does not require key stretching to prevent brute forcing\n(although rate-limiting the token endpoint may still be desirable).","ref":"MyApp.CharonOauth2.Client.html","title":"MyApp.CharonOauth2.Client","type":"module"},{"doc":"Basic changeset.","ref":"MyApp.CharonOauth2.Client.html#changeset/2","title":"MyApp.CharonOauth2.Client.changeset/2","type":"function"},{"doc":"Insert-only changeset.","ref":"MyApp.CharonOauth2.Client.html#insert_only_changeset/2","title":"MyApp.CharonOauth2.Client.insert_only_changeset/2","type":"function"},{"doc":"Returns a new query with the current module as named binding `:charon_oauth2_client`.","ref":"MyApp.CharonOauth2.Client.html#named_binding/0","title":"MyApp.CharonOauth2.Client.named_binding/0","type":"function"},{"doc":"Preload named bindings. Automatically joins using `resolve_binding/2`.","ref":"MyApp.CharonOauth2.Client.html#preload/2","title":"MyApp.CharonOauth2.Client.preload/2","type":"function"},{"doc":"Resolve named bindings that are not present in the query by (left-)joining to the appropriate tables.","ref":"MyApp.CharonOauth2.Client.html#resolve_binding/2","title":"MyApp.CharonOauth2.Client.resolve_binding/2","type":"function"},{"doc":"Bindings / preloads that can be used with `resolve_binding/2` and `preload/2`","ref":"MyApp.CharonOauth2.Client.html#t:resolvable/0","title":"MyApp.CharonOauth2.Client.resolvable/0","type":"type"},{"doc":"","ref":"MyApp.CharonOauth2.Client.html#t:t/0","title":"MyApp.CharonOauth2.Client.t/0","type":"type"},{"doc":"Context to manage clients","ref":"MyApp.CharonOauth2.Clients.html","title":"MyApp.CharonOauth2.Clients","type":"module"},{"doc":"Get a list of all oauth2 clients.","ref":"MyApp.CharonOauth2.Clients.html#all/2","title":"MyApp.CharonOauth2.Clients.all/2","type":"function"},{"doc":"iex> client = insert_test_client!(owner_id: insert_test_user().id)\n    iex> [^client] = Clients.all()\n\n    # can be filtered\n    iex> client = insert_test_client!(owner_id: insert_test_user().id)\n    iex> [^client] = Clients.all(%{owner_id: client.owner_id})\n    iex> [^client] = Clients.all(%{grant_types: \"authorization_code\"})\n    iex> [] = Clients.all(%{owner_id: client.owner_id + 1})","ref":"MyApp.CharonOauth2.Clients.html#all/2-doctests","title":"Doctests - MyApp.CharonOauth2.Clients.all/2","type":"function"},{"doc":"Delete a client.","ref":"MyApp.CharonOauth2.Clients.html#delete/1","title":"MyApp.CharonOauth2.Clients.delete/1","type":"function"},{"doc":"# client must exist\n    iex> {:error, :not_found} = Clients.delete(id: Ecto.UUID.generate())\n\n    # succesfully deletes a client\n    iex> client = insert_test_client!(owner_id: insert_test_user().id)\n    iex> {:ok, _} = Clients.delete([id: client.id])\n    iex> {:error, :not_found} = Clients.delete([id: client.id])","ref":"MyApp.CharonOauth2.Clients.html#delete/1-examples-doctests","title":"Examples / doctests - MyApp.CharonOauth2.Clients.delete/1","type":"function"},{"doc":"Get a single client by one or more clauses, optionally with preloads.\nReturns nil if Client cannot be found.","ref":"MyApp.CharonOauth2.Clients.html#get_by/2","title":"MyApp.CharonOauth2.Clients.get_by/2","type":"function"},{"doc":"iex> client = insert_test_client!(owner_id: insert_test_user().id)\n    iex> %Client{} = Clients.get_by(id: client.id)\n    iex> nil = Clients.get_by(id: Ecto.UUID.generate())\n\n    # preloads things\n    iex> client = insert_test_client!(owner_id: insert_test_user().id)\n    iex> auth = insert_test_authorization!(resource_owner_id: insert_test_user().id, client_id: client.id)\n    iex> insert_test_grant!(authorization_id: auth.id)\n    iex> %{owner: %{id: _}, authorizations: [_]} = Clients.get_by([id: client.id], Client.supported_preloads)","ref":"MyApp.CharonOauth2.Clients.html#get_by/2-doctests","title":"Doctests - MyApp.CharonOauth2.Clients.get_by/2","type":"function"},{"doc":"Insert a new client.","ref":"MyApp.CharonOauth2.Clients.html#insert/1","title":"MyApp.CharonOauth2.Clients.insert/1","type":"function"},{"doc":"# succesfully creates a client with a secret\n    iex> user = insert_test_user()\n    iex> {:ok, client} = insert_test_client(owner_id: user.id)\n    iex> %{secret: <<_::binary>>} = client\n\n    # owner must exist\n    iex> insert_test_client(owner_id: -1) |> errors_on()\n    %{owner: [\"does not exist\"]}\n\n    iex> Clients.insert(%{}) |> errors_on()\n    %{grant_types: [\"can't be blank\"], name: [\"can't be blank\"], owner_id: [\"can't be blank\"], redirect_uris: [\"can't be blank\"], scope: [\"can't be blank\"]}","ref":"MyApp.CharonOauth2.Clients.html#insert/1-examples-doctests","title":"Examples / doctests - MyApp.CharonOauth2.Clients.insert/1","type":"function"},{"doc":"Update a client.","ref":"MyApp.CharonOauth2.Clients.html#update/2","title":"MyApp.CharonOauth2.Clients.update/2","type":"function"},{"doc":"iex> client = insert_test_client!(owner_id: insert_test_user().id)\n    iex> {:ok, updated} = Clients.update(client, %{secret: \"new!\"})\n    iex> false = updated.secret == client.secret\n\n    # secret is randomly generated on update\n    iex> client = insert_test_client!(owner_id: insert_test_user().id)\n    iex> {:ok, updated} = Clients.update([id: client.id], %{secret: \"new!\"})\n    iex> false = updated.secret == client.secret\n    iex> false = updated.secret == \"new!\"\n\n    # scopes must be subset of configured scopes\n    iex> client = insert_test_client!(owner_id: insert_test_user().id)\n    iex> Clients.update([id: client.id], %{scope: ~w(cry)}) |> errors_on()\n    %{scope: [\"must be subset of party, read, write\"]}\n\n    # underlying authrorization scopes are reduced to client's reduced scopes\n    iex> client = insert_test_client!(owner_id: insert_test_user().id, scope: ~w(read write))\n    iex> authorization = insert_test_authorization!(resource_owner_id: insert_test_user().id, client_id: client.id, scope: ~w(read write))\n    iex> {:ok, _} = Clients.update([id: client.id], %{scope: ~w(read)})\n    iex> %{scope: ~w(read)} = Authorizations.get_by(id: authorization.id)\n\n    # id and owner id can't be updated\n    iex> %{id: id, owner_id: owner_id} = insert_test_client!(owner_id: insert_test_user().id)\n    iex> {:ok, %{id: ^id, owner_id: ^owner_id}} = Clients.update([id: id], %{id: Ecto.UUID.generate(), owner_id: -1})","ref":"MyApp.CharonOauth2.Clients.html#update/2-examples-doctests","title":"Examples / doctests - MyApp.CharonOauth2.Clients.update/2","type":"function"},{"doc":"","ref":"MyApp.CharonOauth2.Config.html","title":"MyApp.CharonOauth2.Config","type":"module"},{"doc":"","ref":"MyApp.CharonOauth2.Config.html#get/0","title":"MyApp.CharonOauth2.Config.get/0","type":"function"},{"doc":"","ref":"MyApp.CharonOauth2.Config.html#get_auth_uri/0","title":"MyApp.CharonOauth2.Config.get_auth_uri/0","type":"function"},{"doc":"","ref":"MyApp.CharonOauth2.Config.html#get_secret/0","title":"MyApp.CharonOauth2.Config.get_secret/0","type":"function"},{"doc":"A grant is an (in-progress) Oauth2 flow to obtain auth tokens.\n\nNot every flow needs server-side state to be stored, for example, the implicit grant\nimmediately returns an access token in response to the authorization request, and does\nnot have a second exchange-code-for-token request-response cycle.","ref":"MyApp.CharonOauth2.Grant.html","title":"MyApp.CharonOauth2.Grant","type":"module"},{"doc":"Insert-only changeset - some things (should) never change.","ref":"MyApp.CharonOauth2.Grant.html#insert_only_changeset/2","title":"MyApp.CharonOauth2.Grant.insert_only_changeset/2","type":"function"},{"doc":"Sets the current module name as a camelcase named binding in an Ecto query","ref":"MyApp.CharonOauth2.Grant.html#named_binding/0","title":"MyApp.CharonOauth2.Grant.named_binding/0","type":"function"},{"doc":"Preload named bindings. Automatically joins using `resolve_binding/2`.","ref":"MyApp.CharonOauth2.Grant.html#preload/2","title":"MyApp.CharonOauth2.Grant.preload/2","type":"function"},{"doc":"Resolve named bindings that are not present in the query by (left-)joining to the appropriate tables.","ref":"MyApp.CharonOauth2.Grant.html#resolve_binding/2","title":"MyApp.CharonOauth2.Grant.resolve_binding/2","type":"function"},{"doc":"Bindings / preloads that can be used with `resolve_binding/2` and `preload/2`","ref":"MyApp.CharonOauth2.Grant.html#t:resolvable/0","title":"MyApp.CharonOauth2.Grant.resolvable/0","type":"type"},{"doc":"","ref":"MyApp.CharonOauth2.Grant.html#t:t/0","title":"MyApp.CharonOauth2.Grant.t/0","type":"type"},{"doc":"Context to manage grants","ref":"MyApp.CharonOauth2.Grants.html","title":"MyApp.CharonOauth2.Grants","type":"module"},{"doc":"Get a list of all oauth2 grants.","ref":"MyApp.CharonOauth2.Grants.html#all/2","title":"MyApp.CharonOauth2.Grants.all/2","type":"function"},{"doc":"iex> insert_test_grant!(resource_owner_id: insert_test_user().id)\n    iex> [%Grant{}] = Grants.all()\n\n    # can be filtered\n    iex> grant = insert_test_grant!(resource_owner_id: insert_test_user().id)\n    iex> [%Grant{}] = Grants.all(%{authorization_id: grant.authorization_id})\n    iex> [%Grant{}] = Grants.all(%{code: grant.code})\n    iex> [] = Grants.all(%{authorization_id: grant.authorization_id + 1})","ref":"MyApp.CharonOauth2.Grants.html#all/2-doctests","title":"Doctests - MyApp.CharonOauth2.Grants.all/2","type":"function"},{"doc":"Delete a grant.","ref":"MyApp.CharonOauth2.Grants.html#delete/1","title":"MyApp.CharonOauth2.Grants.delete/1","type":"function"},{"doc":"# grant must exist\n    iex> {:error, :not_found} = Grants.delete(id: -1)\n\n    # succesfully deletes a grant\n    iex> grant = insert_test_grant!(resource_owner_id: insert_test_user().id)\n    iex> {:ok, _} = Grants.delete([id: grant.id])\n    iex> {:error, :not_found} = Grants.delete([id: grant.id])","ref":"MyApp.CharonOauth2.Grants.html#delete/1-examples-doctests","title":"Examples / doctests - MyApp.CharonOauth2.Grants.delete/1","type":"function"},{"doc":"Delete all grants older than the configured `grant_ttl`.","ref":"MyApp.CharonOauth2.Grants.html#delete_expired/0","title":"MyApp.CharonOauth2.Grants.delete_expired/0","type":"function"},{"doc":"iex> valid = insert_test_grant!(resource_owner_id: insert_test_user().id)\n    iex> expired = insert_test_grant!(resource_owner_id: insert_test_user().id)\n    iex> past = DateTime.utc_now() |> DateTime.add(-10)\n    iex> from(t in Grant, where: t.id == ^expired.id) |> Repo.update_all(set: [expires_at: past])\n    iex> Grants.delete_expired()\n    iex> valid_id = valid.id\n    iex> [%{id: ^valid_id}] = Grants.all()","ref":"MyApp.CharonOauth2.Grants.html#delete_expired/0-examples-doctests","title":"Examples / doctests - MyApp.CharonOauth2.Grants.delete_expired/0","type":"function"},{"doc":"Get a single grant by one or more clauses, optionally with preloads.\nReturns nil if Grant cannot be found.","ref":"MyApp.CharonOauth2.Grants.html#get_by/2","title":"MyApp.CharonOauth2.Grants.get_by/2","type":"function"},{"doc":"iex> grant = insert_test_grant!(resource_owner_id: insert_test_user().id)\n    iex> %Grant{} = Grants.get_by(id: grant.id)\n    iex> nil = Grants.get_by(id: grant.id + 1)\n\n    # preloads things\n    iex> grant = insert_test_grant!(resource_owner_id: insert_test_user().id)\n    iex> %{authorization: %{client: %{id: _}}} = Grants.get_by([id: grant.id], Grant.supported_preloads)\n\n    # a grant can be retrieved by its code (actually by the HMAC of its code)\n    iex> %{id: id, code: code} = insert_test_grant!(resource_owner_id: insert_test_user().id)\n    iex> ^id = Grants.get_by(code: code).id","ref":"MyApp.CharonOauth2.Grants.html#get_by/2-doctests","title":"Doctests - MyApp.CharonOauth2.Grants.get_by/2","type":"function"},{"doc":"Insert a new grant","ref":"MyApp.CharonOauth2.Grants.html#insert/1","title":"MyApp.CharonOauth2.Grants.insert/1","type":"function"},{"doc":"# succesfully creates a grant\n    iex> {:ok, _} = insert_test_grant(resource_owner_id: insert_test_user().id)\n\n    iex> Grants.insert(%{}) |> errors_on()\n    %{authorization_id: [\"can't be blank\"], type: [\"can't be blank\"], resource_owner_id: [\"can't be blank\"]}\n\n    # authorization must exist\n    iex> insert_test_grant(resource_owner_id: insert_test_user().id, authorization_id: -1) |> errors_on()\n    %{authorization: [\"does not exist\"]}\n\n    # resource owner must exist and must match the authorization's owner\n    iex> user1 = insert_test_user().id\n    iex> user2 = insert_test_user().id\n    iex> insert_test_grant(resource_owner_id: user1, authorization_id: insert_test_authorization!(resource_owner_id: user2).id) |> errors_on()\n    %{authorization_id: [\"belongs to other resource owner\"]}\n\n    # type must be one of client grant_type's\n    iex> client = insert_test_client!(owner_id: insert_test_user().id, grant_types: ~w(refresh_token))\n    iex> authorization = insert_test_authorization!(resource_owner_id: insert_test_user().id, client_id: client.id)\n    iex> insert_test_grant(authorization_id: authorization.id) |> errors_on()\n    %{type: [\"not supported by client\"]}\n\n    # redirect_uri must be one of client redirect_uri's\n    iex> insert_test_grant(resource_owner_id: insert_test_user().id, redirect_uri: \"https://boom\") |> errors_on()\n    %{redirect_uri: [\"does not match client\"]}\n\n    # redirect_uri is required if client has multiple uris set\n    iex> client = insert_test_client!(owner_id: insert_test_user().id, redirect_uris: ~w(https://a https://b))\n    iex> authorization = insert_test_authorization!(resource_owner_id: insert_test_user().id, client_id: client.id)\n    iex> insert_test_grant(authorization_id: authorization.id, redirect_uri: nil) |> errors_on()\n    %{redirect_uri: [\"can't be blank\"]}\n    iex> insert_test_grant(authorization_id: authorization.id, redirect_uri: \"https://c\") |> errors_on()\n    %{redirect_uri: [\"does not match client\"]}","ref":"MyApp.CharonOauth2.Grants.html#insert/1-examples-doctests","title":"Examples / doctests - MyApp.CharonOauth2.Grants.insert/1","type":"function"},{"doc":"The Oauth2 [authorization endpoint](https://www.rfc-editor.org/rfc/rfc6749#section-3.1).\n\nThis endpoint is meant to be combined with a first-party web client in which a user can grant or deny access.\nSo it does not, by itself, behave as an oauth2 authorization endpoint (for example, it only supports POST requests).\nThe endpoint returns a JSON response with 200 OK and a `redirect_to` parameter,\nthat the web client should then, you know, redirect to.\nThe redirect may also contain an error result.\nHowever, there are some errors that must not result in redirection,\nfor example errors in validating the redirect URI itself.\nSuch errors result in a 400 response with an error description,\nthat the companion web client may show to the user at its discretion.\n\nUser must be logged-in using Charon (at least `Charon.TokenPlugs.verify_token_signature/2` must be called on the conn.)","ref":"MyApp.CharonOauth2.Plugs.AuthorizationEndpoint.html","title":"MyApp.CharonOauth2.Plugs.AuthorizationEndpoint","type":"module"},{"doc":"The endpoint essentially upserts a user's authorization for a client (if the request is valid, of course).\nA subtlety to note is that the endpoint does not ever reduce the authorized scopes, it only expands them.\nThe reason for this is that an Oauth2 client may request multiple access tokens,\nsome of which have a smaller scope than others.\nIf we reduced the scopes to the request scope, the user would need to re-grant permission every time.\n\nIn other words, this endpoint is not intended to manage a user's authorizations.\nIt is probably a good idea for a user to be able to fully revoke or reduce the scope of an authorization.\nFor such purposes, applications can simply add authorization CRUD functionality to their API,\nusing `MyApp.CharonOauth2.Authorizations` functions.","ref":"MyApp.CharonOauth2.Plugs.AuthorizationEndpoint.html#module-behaviour","title":"Behaviour - MyApp.CharonOauth2.Plugs.AuthorizationEndpoint","type":"module"},{"doc":"alias Elixir.MyApp.CharonOauth2.Plugs.AuthorizationEndpoint\n\n    # this endpoint MUST only be useable by the first-party authorizing client!!!\n    forward \"/oauth2/authorize\", AuthorizationEndpoint, config: @config","ref":"MyApp.CharonOauth2.Plugs.AuthorizationEndpoint.html#module-usage","title":"Usage - MyApp.CharonOauth2.Plugs.AuthorizationEndpoint","type":"module"},{"doc":"The Oauth2 [token endpoint](https://www.rfc-editor.org/rfc/rfc6749#section-3.2).\n\nThere is no need to pass the conn through `Plug.Parsers`, because it does so by itself.\nThe endpoint only accepts content type `application/x-www-form-urlencoded` with utf-8 params.\nThe exceptions raised by Plug.Parsers simply bubble up,\nyour application must return appropriate responses for them (Phoenix applications should do so by default).","ref":"MyApp.CharonOauth2.Plugs.TokenEndpoint.html","title":"MyApp.CharonOauth2.Plugs.TokenEndpoint","type":"module"},{"doc":"alias Elixir.MyApp.CharonOauth2.Plugs.TokenEndpoint\n\n    # this endpoint must be public, without any additional authentication requirements\n    forward \"/oauth2/token\", TokenEndpoint, config: @config","ref":"MyApp.CharonOauth2.Plugs.TokenEndpoint.html#module-usage","title":"Usage - MyApp.CharonOauth2.Plugs.TokenEndpoint","type":"module"},{"doc":"Confidential clients must authenticate using their client secret (and public clients may do so too\nalthough there would be little point in it if they can't keep their secret, you know, secret).\n\nHTTP Basic authentication takes precedence over req body credentials.\nAlthough [the spec](https://datatracker.ietf.org/doc/html/rfc6749#section-2.3)\nsays clients must not use more than one auth method,\nit doesn't say auth servers should reject such requests,\nso we simply decided that HTTP Basic takes priority,\nbecause the same spec says that HTTP Basic, at least, *must* be supported\nfor clients that identify using a client password.\n\nIF HTTP Basic auth is used, an authentication failure results in a 401 response with\nheader \"www-authenticate\" set to \"Basic\", instead of a \"normal\" 400-with-JSON error response.","ref":"MyApp.CharonOauth2.Plugs.TokenEndpoint.html#module-client-authentication","title":"Client authentication - MyApp.CharonOauth2.Plugs.TokenEndpoint","type":"module"},{"doc":"The Oauth 2.1 [draft spec](https://www.ietf.org/archive/id/draft-ietf-oauth-v2-1-07.html) recommends\nenforcing PKCE for the authorization_code grant under all circumstances. That is what we do.","ref":"MyApp.CharonOauth2.Plugs.TokenEndpoint.html#module-authorization-code-with-proof-key-for-code-exchange-pkce-enforced","title":"Authorization code with Proof Key for Code Exchange (PKCE) enforced - MyApp.CharonOauth2.Plugs.TokenEndpoint","type":"module"},{"doc":"Tokens handed out by this endpoint are backed by a server-side session (that is only loaded on refresh, by default).\nThe session type is `:oauth2`, separating these sessions from other sessions that the user may have in `Charon`.\nThe purpose of this separation is to be able to call delete-all and not drop oauth2-client connections,\nor the other way around.","ref":"MyApp.CharonOauth2.Plugs.TokenEndpoint.html#module-sessions","title":"Sessions - MyApp.CharonOauth2.Plugs.TokenEndpoint","type":"module"},{"doc":"Insert test values for writing tests. Uses default values described below.\n\nAll functions take an `overrides` parameter that can be used to override the default values.\nIn order to set your own defaults, you can use the `CharonOauth2.Config` field `:test_seed_defaults`.","ref":"MyApp.CharonOauth2.TestSeeds.html","title":"MyApp.CharonOauth2.TestSeeds","type":"module"},{"doc":"Inserts an authorization using `Elixir.MyApp.CharonOauth2.Authorizations.insert/1`.\n\nThe default values can be overridden using `overrides`. These are the default values:\n\n```\n%{scope: [\"read\"]}\n```","ref":"MyApp.CharonOauth2.TestSeeds.html#insert_test_authorization/1","title":"MyApp.CharonOauth2.TestSeeds.insert_test_authorization/1","type":"function"},{"doc":"Inserts an authorization using `Elixir.MyApp.CharonOauth2.Authorizations.insert/1`. Raises on failure.\n\nThe default values can be overridden using `overrides`. These are the default values:\n\n```\n%{scope: [\"read\"]}\n```","ref":"MyApp.CharonOauth2.TestSeeds.html#insert_test_authorization!/1","title":"MyApp.CharonOauth2.TestSeeds.insert_test_authorization!/1","type":"function"},{"doc":"Inserts a client using `Elixir.MyApp.CharonOauth2.Clients.insert/1`.\n\nThe default values can be overridden using `overrides`. These are the default values:\n\n```\n%{\n  description: \"Incredibly innovative application that definitely treats your data well.\",\n  grant_types: [\"authorization_code\"],\n  name: \"MyClient\",\n  redirect_uris: [\"https://mysite.tld\"],\n  scope: [\"read\"]\n}\n```","ref":"MyApp.CharonOauth2.TestSeeds.html#insert_test_client/1","title":"MyApp.CharonOauth2.TestSeeds.insert_test_client/1","type":"function"},{"doc":"Inserts a client using `Elixir.MyApp.CharonOauth2.Clients.insert/1`. Raises on failure.\n\nThe default values can be overridden using `overrides`. These are the default values:\n\n```\n%{\n  description: \"Incredibly innovative application that definitely treats your data well.\",\n  grant_types: [\"authorization_code\"],\n  name: \"MyClient\",\n  redirect_uris: [\"https://mysite.tld\"],\n  scope: [\"read\"]\n}\n```","ref":"MyApp.CharonOauth2.TestSeeds.html#insert_test_client!/1","title":"MyApp.CharonOauth2.TestSeeds.insert_test_client!/1","type":"function"},{"doc":"Inserts a grant using `Elixir.MyApp.CharonOauth2.Grants.insert/1`.\n\nThe default values can be overridden using `overrides`. These are the default values:\n\n```\n%{redirect_uri: \"https://mysite.tld\", type: \"authorization_code\"}\n```","ref":"MyApp.CharonOauth2.TestSeeds.html#insert_test_grant/1","title":"MyApp.CharonOauth2.TestSeeds.insert_test_grant/1","type":"function"},{"doc":"Inserts a grant using `Elixir.MyApp.CharonOauth2.Grants.insert/1`. Raises on failure.\n\nThe default values can be overridden using `overrides`. These are the default values:\n\n```\n%{redirect_uri: \"https://mysite.tld\", type: \"authorization_code\"}\n```","ref":"MyApp.CharonOauth2.TestSeeds.html#insert_test_grant!/1","title":"MyApp.CharonOauth2.TestSeeds.insert_test_grant!/1","type":"function"},{"doc":"# CharonOauth2\n\nCharonOauth2 is a child package of [Charon](https://github.com/weareyipyip/charon) that adds Oauth2 authorization server capability to a Charon-secured application. Charon is an auth framework for Elixir aimed primarily at securing APIs. If you simply add CharonOauth2 to an existing application, you will probably end up with an API that is both the Oauth2 authorization server and resource server. That is perfectly fine, just be careful designing and enforcing your scopes.\n\nBecause Charon focuses on securing APIs, CharonOauth2 does not include an \"authorize app X to use your Y account\" page. Such a page is necessary for Oauth2, and adding it is left to the client application using whatever stack it chooses. CharonOauth does include an endpoint plug that does almost all of the heavy lifting of such a page, so that CharonOauth2 can be easily used without having to dive into the Oauth2 spec. This makes CharonOauth2 a little less of a batteries-included solution, but since you probably have a web app already, or a stack preference, or obnoxious developers or whatever, this shouldn't be too much of an issue in practice.\n\nCharonOauth2 implements recommendations from the [Oauth 2.1 draft spec](https://www.ietf.org/archive/id/draft-ietf-oauth-v2-1-07.html), such as enforcing PKCE with the authorization code grant for all clients.","ref":"readme.html","title":"CharonOauth2","type":"extras"},{"doc":"<!-- TOC -->\n\n- [Charon](#Charon)\n  - [Table of contents](#table-of-contents)\n  - [Features](#features)\n  - [Documentation](#documentation)\n  - [How to use](#how-to-use)\n    - [Installation](#installation)\n    - [Set up Charon](#set-up-charon)\n    - [Configuration](#configuration)\n    - [Migrations](#migrations)\n    - [Create your CharonOauth2 module](#create-your-charonoauth2-module)\n    - [Add Oauth2 routes](#add-oauth2-routes)\n    - [Add an authorization page](#add-an-authorization-page)\n      - [Add API endpoints needed for authorization page](#add-api-endpoints-needed-for-authorization-page)\n    - [Restrict third-party application access using scopes](#restrict-third-party-application-access-using-scopes)\n      - [What are scopes?](#what-are-scopes)\n      - [Enforcing scopes](#enforcing-scopes)\n    - [Managing authorizations and clients](#managing-authorizations-and-clients)\n    - [Testing](#testing)\n\n<!-- /TOC -->","ref":"readme.html#table-of-contents","title":"Table of contents - CharonOauth2","type":"extras"},{"doc":"- Oauth 2.1 authorization server implementation supporting authorization_code and refresh_token grants.\n- Out-of-the-box database migrations, models, and contexts.\n- Simple configuration with sane defaults.\n- Revokable refresh tokens thanks to Charon's session store.\n- Symmetric or asymmetric token signatures thanks to Charon's token factory.\n- Safe storage of sensitive data using encryption (`CharonOauth2.Types.Encrypted`) or HMAC (`CharonOauth2.Types.Hmac`) when appropriate.\n- Small number of dependencies.","ref":"readme.html#features","title":"Features - CharonOauth2","type":"extras"},{"doc":"Documentation can be found at [https://hexdocs.pm/charon_oauth2](https://hexdocs.pm/charon_oauth2).","ref":"readme.html#documentation","title":"Documentation - CharonOauth2","type":"extras"},{"doc":"","ref":"readme.html#how-to-use","title":"How to use - CharonOauth2","type":"extras"},{"doc":"The package can be installed by adding `charon_oauth2` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:charon, \"~> 3.1\"},\n    {:charon_oauth2, \"~> 0.0\"}\n  ]\nend\n```","ref":"readme.html#installation","title":"Installation - CharonOauth2","type":"extras"},{"doc":"Set up [Charon](https://github.com/weareyipyip/charon) first using its readme.","ref":"readme.html#set-up-charon","title":"Set up Charon - CharonOauth2","type":"extras"},{"doc":"Configuration is easy. You simply add the `CharonOauth2` configuration as optional module configuration for Charon.\nAdd your repo, resource owner (user) schema, and scopes. We will discuss scopes in more detail [later](#restrict-third-party-application-access-using-scopes). For now, you can add an empty list.\n\n```elixir\nCharon.Config.from_enum(\n  ...,\n  optional_modules: %{\n    CharonOauth2 => %{\n      repo: MyApp.Repo,\n      resource_owner_schema: MyApp.User,\n      scopes: []\n    }\n  }\n)\n```\n\nFor all config options, take a look at `CharonOauth2.Config`.","ref":"readme.html#configuration","title":"Configuration - CharonOauth2","type":"extras"},{"doc":"To create the required database tables, you can use the included migration helper `CharonOauth2.Migration`.\n\n```bash\nmix ecto.gen.migration charon_oauth2_models\n```\n\n```elixir\ndefmodule MyApp.Repo.Migrations.CharonOauth2Models do\n  use Ecto.Migration\n\n  @my_charon_config MyApp.Charon.get_config()\n\n  def change, do: CharonOauth2.Migration.change(@my_charon_config)\nend\n```","ref":"readme.html#migrations","title":"Migrations - CharonOauth2","type":"extras"},{"doc":"To provide you with models, contexts and plugs, CharonOauth2 generates several modules for you.\nCreate a CharonOauth2 module, simply passing in the Charon config:\n\n```elixir\ndefmodule MyApp.CharonOauth2 do\n  use CharonOauth2, MyApp.Charon.get_config()\nend\n```\n\nNow you will have the following additional modules:\n\n- `MyApp.CharonOauth2.Authorization`\n- `MyApp.CharonOauth2.Authorizations`\n- `MyApp.CharonOauth2.Client`\n- `MyApp.CharonOauth2.Clients`\n- `MyApp.CharonOauth2.Grant`\n- `MyApp.CharonOauth2.Grants`\n- `MyApp.CharonOauth2.Plugs.AuthorizationEndpoint`\n- `MyApp.CharonOauth2.Plugs.TokenEndpoint`\n\nNow you only have to update your user schema:\n\n```elixir\ndefmodule MyApp.User do\n  use Ecto.Schema\n\n  alias MyApp.CharonOauth2.{Authorization, Client}\n\n  schema \"users\" do\n    ...\n\n    has_many :oauth2_authorizations, Authorization, foreign_key: :resource_owner_id\n    has_many :oauth2_clients, Client, foreign_key: :owner_id\n  end\nend\n```","ref":"readme.html#create-your-charonoauth2-module","title":"Create your CharonOauth2 module - CharonOauth2","type":"extras"},{"doc":"Oauth2 requires two important endpoints, the authorization endpoint and the token endpoint.\nYou can add both by forwarding to the two \"endpoint plugs\":\n\n```elixir\ndefmodule MyAppWeb.Router do\n  use MyAppWeb, :router\n  alias MyApp.CharonOauth2.Plugs.{AuthorizationEndpoint, TokenEndpoint}\n  import Charon.TokenPlugs\n\n  @my_charon_config MyApp.Charon.get_config()\n\n  ...\n\n  scope \"/api\" do\n    pipe_through [:api ,:valid_access_token]\n\n    # auth endpoint MUST only be accessible by Charon-authenticated users\n    forward \"/oauth2/authorize\", AuthorizationEndpoint, config: @my_charon_config\n  end\n\n  scope \"/api\" do\n    # token endpoint does its own request body parsing and requires no authentication\n    forward \"/oauth2/token\", TokenEndpoint, config: @my_charon_config\n  end\nend\n```","ref":"readme.html#add-oauth2-routes","title":"Add Oauth2 routes - CharonOauth2","type":"extras"},{"doc":"The authorization page receives the authorization request from third-party apps (Oauth2 clients). Its purpose is to allow the user to grant or deny permission to access or use (parts of) the user's account. This access is limited by scopes.\n\n`MyApp.CharonOauth2.AuthorizationEndpoint` verifies everything that needs verifying, so your implementation does not necessarily have to concern itself with validating query params. However, to provide a nice UX, it will have to make sure that:\n\n1. `client_id` is present\n1. the user is logged-in (redirect to login page otherwise)\n\nThe page must do the following:\n\n1. Fetch an existing authorization for the client by the logged-in user from the API.\n1. Fetch the client from the API.\n1. Fetch all defined scopes with their descriptions from the API.\n1. Determine which scopes are requested:\n   - IF the `scope` query parameter is set, split it on whitespace \" \".\n   - OR the fetched client's configured `scope`.\n1. Compare the requested scopes to the already-authorized scopes (if any).\n1. IF there are yet-to-be-authorized scopes, show them to the user and ask for permission.\n1. Call the authorization endpoint with all query params in the request body plus `\"permission_granted\":  ` (this should be `true` if the user has already granted permission for all requested scopes).\n1. Process the response\n   - IF 200 OK and `redirect_to`, then redirect to the provided link. This redirect may also contain an error response.\n   - IF 400 Bad Request and `errors`, show the user an error message and _don't redirect_ (this basically only happens if something is wrong with request parameters that would make it unsafe to redirect to the client, for example, the `redirect_uri` does not match the one configured for the client).\n\n#### React example\n\nA React example can be found [here](./example-auth-page.md).\n\n#### Add API endpoints needed for authorization page\n\nYou will need the following (or similar in GraphQL or something):\n\n- `GET /api/scopes` (public) all defined application scopes, and probably their (default) descriptions, e.g. `{\"profile:read\": \"View your profile details like your name and email.\"}`.\n- `GET /api/oauth2_clients/:id` (public) Oauth2 client by ID. Be careful _not_ to render the client secret here!\n- `GET /api/my/oauth2_authorizations/:client_id` (private) current user's existing authorization for the specified client.\n\nImplementations for these endpoints are trivial and are not shown.","ref":"readme.html#add-an-authorization-page","title":"Add an authorization page - CharonOauth2","type":"extras"},{"doc":"You are now basically done as far as adding `CharonOauth2` to your app is concerned. Hurray!\nHowever, you actually still have to do the most difficult part :see_no_evil:,\nwhich is restricting third-party access to your API using scopes\n(technically, the part of it that functions as an Oauth2 resource server).\nFor reasons explained below, this is something that cannot be abstracted away in a library.\n\n#### What are scopes?\n\nScopes are not the easiest concept to grasp in the first place.\nThey are best thought of as permissions for applications, as opposed to permissions for users.\nFor an operation to be authorized, both the application and the user must be authorized to perform it.\n\nFor example, let's imagine your app can open a building's door.\nA user is authorized to open a door if they belong to the building (user permissions)\nHowever, you don't want just any third-party app to open a door on behalf of your user,\nexcept when the user has explicitly granted permission to that app (application permissions).\nThat is why you want to restrict open-door capability to apps with scope `door:open`, for example.\n\n_So in an Oauth2-enabled app, user permissions (roles etc) can't replace scopes, and scopes can't replace user permissions._\n\nIt is not always straightforward which scopes to define for your app,\nnor is it always simple which scopes should be needed for which operations.\nDoes `door:open` imply `door:read`, in other words, are scopes hierarchical?\nIf your open-door endpoint returns 204 No Content, maybe requiring `door:read` is not a good fit, after all, what data are you reading?\nOn the other hand, if the endpoint returns the configuration data of the opened door,\nyou are effectively reading the door's configuration by opening it, and\nrequiring `door:read` too seems logical.\nSo `door:open` does not necessarily imply `door:read` and it really depends on the application.\nIt's probably best if scopes are not hierarchical and you simply require multiple scopes for the same operation where appropriate,\nnot in the last place to preserve your own sanity.\n\nAnother problem arises if your API returns other users' data.\nFor example, if your application provides a `GET /my/building/residents` endpoint,\nyou may run into privacy or even legal issues if you let third-party apps access that\nendpoint. Your users may have agreed somewhere that their co-residents can see their names,\nand that your application knows who live together in the same building. So without\nthird-party access, everything is fine.\n\nHowever, if you then add Oauth2 third-party app access, a user grants permission\nto such an app to use their account, and that app can then see other user's names,\na situation has arisen in which those other users have NOT granted any permissions to the third-party app,\nbut it has some of their data anyway.\n\nSo design your API and scopes well.\n\n##### Recommendations\n\nYou should probably never allow third-party applications to do the following things, to prevent privilege escalation:\n\n- Read, create or update Oauth2 grants, authorizations or clients.\n- Update a user's login credentials like passwords or MFA methods.\n- Read, create or update a user's push tokens.\n- Read, create or update a user's (non-Oauth2) sessions.\n- Access highly privileged accounts like application-wide admin accounts, if you have those.\n- Read other users' profiles.\n- Refresh tokens using an existing Charon session controller.\n\nIt is especially important to be aware of this if your use of CharonOauth2 creates a combined authorization- and resource server.\nYou can (and should) use scopes to enforce these restrictions.\n\n#### Enforcing scopes\n\nThis is the easy part, simply verify the scope claim in a token:\n\n```elixir\ndefmodule MyAppWeb.Router do\n  use MyAppWeb, :router\n  alias MyApp.CharonOauth2.Plugs.{AuthorizationEndpoint, TokenEndpoint}\n  import Charon.TokenPlugs\n\n  @my_charon_config MyApp.Charon.get_config()\n\n  ...\n\n  pipeline :restrict_authorization_access do\n    # the scope claim is an ordset (see :ordsets)\n    # this enforces that *both* scopes are present in the token claim\n    plug :verify_token_ordset_claim_contains, {\"scope\", :ordsets.from_list(~w(authorization:write grant:write))}\n    plug :verify_no_auth_error, &MyAppWeb.charon_error_handler/2\n  end\n\n  scope \"/api\" do\n    pipe_through [:api ,:valid_access_token, :restrict_authorization_access]\n\n    # auth endpoint MUST only be accessible by Charon-authenticated users\n    # AND only by privileged, first-party applications\n    # that have access to scopes authorization:write and grant:write (or similar)\n    forward \"/oauth2/authorize\", AuthorizationEndpoint, config: @my_charon_config\n  end\n\n  scope \"/api\" do\n    # token endpoint does its own parsing and requires no authentication\n    forward \"/oauth2/token\", TokenEndpoint, config: @my_charon_config\n  end\nend\n```\n\nYou can enforce scopes to separate first-party and third-party clients.\n\nTo make sure that first-party clients have scopes that third-party clients don't have,\nyou can simply use an existing session controller\n(for example, the one you created for [Charon](https://github.com/weareyipyip/charon))\nand grant all scopes to tokens handed out by it.\nThis is the easiest and recommended way, because all of the nice things like session management\nwill stay straightforward.\n\nAlternatively, you can go \"full Oauth2\", adding your own first-party client as an Oauth2 client that\nhas access to scopes that other clients don't, and use that for your own applications.\nIn that case you can throw away your \"normal\" Charon session controller.\nAlso, you probably want to \"pre-authorize\" each user for your own client.","ref":"readme.html#restrict-third-party-application-access-using-scopes","title":"Restrict third-party application access using scopes - CharonOauth2","type":"extras"},{"doc":"To round off, you may wish to add the ability for users to register their own Oauth2 clients (be the owner of a third-party app) or manage their authorizations. These require simple CRUD operations on `MyApp.CharonOauth2.Clients` and `MyApp.CharonOauth2.Authorizations`, respectively. Implementing controllers or GraphQL resolvers for those operations is entrusted to the reader's capabilities ;)","ref":"readme.html#managing-authorizations-and-clients","title":"Managing authorizations and clients - CharonOauth2","type":"extras"},{"doc":"If you wish to write tests that involve CharonOauth2 models, you can use the utility functions in\n`MyApp.CharonOauth2.TestSeeds` to insert test seeds.","ref":"readme.html#testing","title":"Testing - CharonOauth2","type":"extras"},{"doc":"# Example authorization page (React)\n\n```tsx\nimport React from \"react\";\n\nimport ApiService from \"@apiService\";\nimport { Oauth2Client } from \"@types\";\nimport { useAppContext } from \"@global/context\";\nimport { Button } from \"@components/atoms/Button/Component\";\n\nexport default function AuthorizeOauth2(): JSX.Element {\n  let params = Object.fromEntries(\n    new URLSearchParams(window.location.search).entries()\n  );\n\n  const { setErrorMessage } = useAppContext();\n\n  const [scopeDescriptions, setScopeDescriptions] = React.useState<{\n    [key: string]: string;\n  }>({});\n  const [authorizedScopes, setAuthorizedScopes] = React.useState ([]);\n  const [newScopes, setNewScopes] = React.useState ([]);\n  const [client, setClient] = React.useState (null);\n\n  React.useEffect((): void => {\n    const clientId = params.client_id;\n    const scopes = params.scope?.split(\" \") ?? [];\n\n    if (clientId) {\n      fetchData(clientId, scopes).catch((error) => {\n        setErrorMessage(\"Something went wrong\");\n        throw error;\n      });\n    }\n  }, [params]);\n\n  const fetchData = async (\n    clientId: string,\n    scopes: string[]\n  ): Promise  => {\n    const [\n      {\n        data: {\n          data: { oauth2_client: client },\n        },\n      },\n      {\n        data: {\n          data: { oauth2_authorizations: authorizations },\n        },\n      },\n      {\n        data: {\n          data: { token_scopes: scopeDescriptions },\n        },\n      },\n    ] = await Promise.all([\n      ApiService.oauth2Clients.show(clientId),\n      ApiService.my.oauth2Authorizations.list({ client_id: clientId }),\n      ApiService.tokenScopes.list(),\n    ]);\n\n    // If specific scopes are requested, we'll use those. Otherwise, default to all client scopes.\n    const requestedScopes = scopes.length > 0 ? scopes : client.scopes;\n    const authorizedScopes = authorizations?.[0]?.scope ?? [];\n    const newScopes = requestedScopes.filter(\n      (s) => !authorizedScopes.includes(s)\n    );\n\n    if (newScopes.length === 0) {\n      // If all requested scopes are already authorized, we can just re-authorize on behalf of the user.\n      await authorize(true);\n      return;\n    }\n\n    setClient(client);\n    setScopeDescriptions(scopeDescriptions);\n    setAuthorizedScopes(authorizedScopes);\n    setNewScopes(newScopes);\n  };\n\n  const authorize = async (grantPermission: boolean): Promise  => {\n    try {\n      const {\n        data: { redirect_to: redirectTo },\n      } = await ApiService.oauth2.authorize({\n        ...params,\n        permission_granted: grantPermission,\n      });\n      window.location.replace(redirectTo);\n    } catch (error: any) {\n      setErrorMessage(JSON.stringify(error.response.data));\n    }\n  };\n\n  if (!client) {\n    return  Loading... ;\n  } else {\n    return (\n       \n         \n           \n             \n              The following app wishes to use your account on your behalf:{\" \"}\n               {client.name}  - {client.description}\n             \n             \n\n            {authorizedScopes.length > 0 && (\n              <>\n                 You have already granted {client.name} permission to: \n                 \n                  {authorizedScopes.map((scope: string, i: number) => (\n                     {scopeDescriptions[scope] ?? scope} \n                  ))}\n                 \n              </>\n            )}\n\n             In addition, {client.name} wants to: \n             \n              {newScopes.map((scope: string, i: number) => (\n                 {scopeDescriptions[scope] ?? scope} \n              ))}\n             \n           \n         \n         \n          Confirm\n         \n         \n          Deny\n         \n       \n    );\n  }\n}\n```","ref":"example-auth-page.html","title":"Example authorization page (React)","type":"extras"}],"producer":{"name":"ex_doc","version":"0.35.1"},"proglang":"elixir"}